#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <arpa/inet.h>
#include <curl/curl.h>
#include <vector>
#include "metacube.h"

using namespace std;

class Input {
public:
	Input();
	void curl_callback(char *ptr, size_t bytes);

private:
	void process_block(const char *data, uint32_t size, uint32_t flags);
	void drop_pending_data(size_t num_bytes);

	// Data we have received but not fully processed yet.
	vector<char> pending_data;

	// If <pending_data> starts with a Metacube header,
	// this is true.
	bool has_metacube_header;
};

Input::Input()
	: has_metacube_header(false)
{
}
	
void Input::curl_callback(char *ptr, size_t bytes)
{
	pending_data.insert(pending_data.end(), ptr, ptr + bytes);

	for ( ;; ) {
		// If we don't have enough data (yet) for even the Metacube header, just return.
		if (pending_data.size() < sizeof(metacube_block_header)) {
			return;
		}

		// Make sure we have the Metacube sync header at the start.
		// We may need to skip over junk data (it _should_ not happen, though).
		if (!has_metacube_header) {
			char *ptr = static_cast<char *>(
				memmem(pending_data.data(), pending_data.size(),
				       METACUBE_SYNC, strlen(METACUBE_SYNC)));
			if (ptr == NULL) {
				// OK, so we didn't find the sync marker. We know then that
				// we do not have the _full_ marker in the buffer, but we
				// could have N-1 bytes. Drop everything before that,
				// and then give up.
				drop_pending_data(pending_data.size() - (strlen(METACUBE_SYNC) - 1));
				return;
			} else {
				// Yay, we found the header. Drop everything (if anything) before it.
				drop_pending_data(ptr - pending_data.data());
				has_metacube_header = true;

				// Re-check that we have the entire header; we could have dropped data.
				if (pending_data.size() < sizeof(metacube_block_header)) {
					return;
				}
			}
		}

		// Now it's safe to read the header.
		metacube_block_header *hdr = reinterpret_cast<metacube_block_header *>(pending_data.data());	
		assert(memcmp(hdr->sync, METACUBE_SYNC, sizeof(hdr->sync)) == 0);
		uint32_t size = ntohl(hdr->size);
		uint32_t flags = ntohl(hdr->flags);

		// See if we have the entire block. If not, wait for more data.
		if (pending_data.size() < sizeof(metacube_block_header) + size) {
			return;
		}

		process_block(pending_data.data(), size, flags);

		// Consume this block. This isn't the most efficient way of dealing with things
		// should we have many blocks, but these routines don't need to be too efficient
		// anyway.
		pending_data.erase(pending_data.begin(), pending_data.begin() + sizeof(metacube_block_header) + size);
	}
}
		
void Input::process_block(const char *data, uint32_t size, uint32_t flags)
{
	// TODO: treat it right here
	printf("Block: %d bytes, flags=0x%x\n", size, flags);
}
			
void Input::drop_pending_data(size_t num_bytes)
{
	if (num_bytes == 0) {
		return;
	}
	fprintf(stderr, "Warning: Dropping %lld junk bytes from stream, maybe it is not a Metacube stream?\n",
		(long long)num_bytes);
	pending_data.erase(pending_data.begin(), pending_data.begin() + num_bytes);
}

size_t curl_callback(char *ptr, size_t size, size_t nmemb, void *userdata)
{
	Input *input = static_cast<Input *>(userdata);
	size_t bytes = size * nmemb;
	input->curl_callback(ptr, bytes);	
	return bytes;
}

int main(int argc, char **argv)
{
	Input input;
	CURL *curl = curl_easy_init();
	curl_easy_setopt(curl, CURLOPT_URL, "http://gruessi.zrh.sesse.net:4013/");
	// curl_easy_setopt(curl, CURLOPT_URL, "http://www.google.com/");
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curl_callback);
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, &input);
	curl_easy_perform(curl);
}
